//
// csvrobot.cpp: CSV content manipulator.
//
// (C) Copyright 2016 TinyIT
// Written by Pawel Tatera
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; version 3
// of the License.
//


#include <iostream>
#include <cstdlib>
#include <cstring>
#include <vector>
#include <list>

#include "scsv.h"
#include "csvrobot_ver.h"

using namespace std;


// Structure used in main program function to control its behaviour.
typedef struct _APP_CFG {
  const char *const namePtr;
  const char *const longNamePtr;
  bool request;
  bool needsParam;
  union {
    const char *argC;
    int argI;
  };
} APP_CFG;


static APP_CFG * get_cfg_ptr(const APP_CFG *, const char *const);
static int strchrcnt(const char * const sPtr, char);
static void cstr_to_ushort_array(const char * const sPtr, char, unsigned short *);
static void usage();
static void app_info();


int main(const int argc, const char *argv[]){
  union {
    int i;
    bool b;
  };
  char cc;
  unsigned short *colsList, *quoteList, *unquoteList;
  scsv csv;

  // Required params for the program to work.
  // Provide a short and optionally long name or if only long name should be
  // available, write it in place of short name!
  APP_CFG appConfig[] = {
    { "-i", "--in", 0, 1, NULL },
    { "-o", "--out", 0, 1, NULL },
    { "-a", NULL, 0, 0, NULL },
    { "-c", NULL, 0, 1, NULL },
    { "-e", NULL, 0, 1, NULL },
    { "-E", NULL, 0, 1, NULL },
    { "-q", NULL, 0, 1, NULL },
    { "-Q", NULL, 0, 1, NULL },
    { "-s", NULL, 0, 1, NULL },
    { "-S", NULL, 0, 1, NULL },
    { "-h", "--help", 0, 0, 0 },
    // { "-v", NULL, 0, 0, 0 },
    { "--version", NULL, 0, 0, 0 },
    { NULL, NULL, 0, 0, 0 } // IMPORTANT!
  }, *cfgP;

  // Based on argc/v input, fill the APP_CFG actions from above appConfig.
  // i=1: argv[0] is a program name
  for(i=1; i<argc; i++){
    cfgP = get_cfg_ptr(appConfig, argv[i]);
    //if(cfgP && (cfgP->namePtr || cfgP->longNamePtr)){
    if(cfgP->namePtr){
      if(cfgP->needsParam){
        // TODO
        // Test if argv[i+1] is really a valid param!
        // ... But it had to be simple! :|
        cfgP->argC = argv[i+1];
        i++;
      }
      cfgP->request = 1;
    } 
    else { usage(); return 0; }
  }

  //
  // Actions that force the program to stop without doing anything more
  //
  cfgP = get_cfg_ptr(appConfig, "-h");
  if(cfgP->request){ usage(); return 0; }

  cfgP = get_cfg_ptr(appConfig, "--version");
  if(cfgP->request){ app_info(); return 0; }

  //
  // Provide startup configuration for libscsv instance.
  //

  // 0-based list of columns id-s to read from each csv line
  // Make sure get_cfg_ptr will always find a match.
  cfgP = get_cfg_ptr(appConfig, "-c");
  if(cfgP->argC){
    // count ',' chars in string to get the number of columns to read
    i=strchrcnt(cfgP->argC, ',');
    if(i >= 1)
      colsList = new (std::nothrow) unsigned short [i];
      if(colsList){
        cstr_to_ushort_array(cfgP->argC, ',', colsList);
        csv.read_cols(i, colsList);
      } else 
        cout
	  << "Warning! Can't allocate memory for provided columns list (-c)."
	  << endl;
  } else colsList = NULL;

  // 0-based list of columns id-s to remove quote character.
  // Make sure get_cfg_ptr will always find a match.
  cfgP = get_cfg_ptr(appConfig, "-e");
  if(cfgP->argC){
    // count ',' chars in string to get the number of columns to read
    i=strchrcnt(cfgP->argC, ',');
    if(i >= 1)
      unquoteList = new (std::nothrow) unsigned short [i];
      if(unquoteList){
        cstr_to_ushort_array(cfgP->argC, ',', unquoteList);
        csv.unquote_cols(i, unquoteList);
      } else 
        cout
	  << "Warning! Can't allocate memory for provided columns list (-e)."
	  << endl;
  } else unquoteList = NULL;

  // 0-based list of columns id-s to enclose between quote character.
  // Make sure get_cfg_ptr will always find a match.
  cfgP = get_cfg_ptr(appConfig, "-E");
  if(cfgP->argC){
    // count ',' chars in string to get the number of columns to read
    i=strchrcnt(cfgP->argC, ',');
    if(i >= 1)
      quoteList = new (std::nothrow) unsigned short [i];
      if(quoteList){
        cstr_to_ushort_array(cfgP->argC, ',', quoteList);
        csv.quote_cols(i, quoteList);
      } else 
        cout
	  << "Warning! Can't allocate memory for provided columns list (-E)."
	  << endl;
  } else quoteList = NULL;

  // field separator to use
  cfgP = get_cfg_ptr(appConfig, "-s");
  if(cfgP->argC) csv.set_iseparator(* cfgP->argC);
  else csv.set_iseparator(',');
  
  cfgP = get_cfg_ptr(appConfig, "-S");
  if(cfgP->argC) csv.set_oseparator(* cfgP->argC);
  else csv.set_oseparator(',');

  cfgP = get_cfg_ptr(appConfig, "-q");
  if(cfgP->argC) csv.set_unquote(* cfgP->argC);
  else csv.set_unquote(0);
 
  cfgP = get_cfg_ptr(appConfig, "-Q");
  if(cfgP->argC) csv.set_quote(* cfgP->argC);
  else csv.set_quote(0);
 
  cfgP = get_cfg_ptr(appConfig, "-i");
  if(cfgP->argC)
    i = csv.read(cfgP->argC);
  else
    i = csv.read(cin);

  // Clean stuff after reading.
  if(colsList) delete [] colsList;
  if(unquoteList) delete [] unquoteList;

  cfgP = get_cfg_ptr(appConfig, "-o");
  if(cfgP->argC){
    // Append mode requested?
    b = (get_cfg_ptr(appConfig, "-a"))->request;
    i = csv.save(cfgP->argC, b); 
  } else
    i = csv.save(cout);

  // Clean stuff afteer writing.
  if(quoteList) delete [] quoteList;

  return i;
}


//
// Supportng functions
//


// Say how many times a character appears in a string.
int strchrcnt(const char *sPtr, char character){
  int i=0;

  if(sPtr){
    const char *p = sPtr;
    while(p){
      i++;
      p = strchr(p, character);
      if(p) p++; // DON'T INCREASE THE NULL POINTER!
    }
  }

  return i;
}

// Return a pointer to the one of cfg elements requested with namePtr.
APP_CFG * get_cfg_ptr(const APP_CFG * cfg, const char *const namePtr){

  while(cfg->namePtr){
    if( strcmp(namePtr, cfg->namePtr)==0) break;

    if(cfg->longNamePtr)
      if( strcmp(namePtr, cfg->longNamePtr)==0 ) break;

    cfg++;
  }

  return (APP_CFG *) cfg; 
}


// Convert cstring into an array of unsigned short-s.
// Use strchrcnt to get the proper size of list.
void cstr_to_ushort_array(const char *cStrP, char c,
  unsigned short *list){

  const char separator[] = { c, 0 };
  char *t = strtok((char*)cStrP, separator);

  while(t){
    *list = (unsigned short) atoi(t);
    list++;
    t = strtok(NULL, separator);
  }
}


//
// Less important things
//

void app_info(){
  cout << EXE_STR_FileDescription << " version " << EXE_STR_V_FILE << endl
    << "Written by Pawel Tatera" << endl
    << EXE_STR_LegalCopyright << " http://www.tinyit.pl" << endl;
}


void usage(){
cout << "Usage:" << endl
<< " csvrobot [ -i in_file ] [ -o out_file ] [ -c cols_list ] [ -s separator ]"
<< endl << "\t[ -q char -e cols_list ] [ -Q char -E cols_list ]"
<< endl << endl
<< " Options: " << endl
<< "  -i --in    input file" << endl
<< "  -o --out   output file" << endl
<< "  -a         enable append mode" << endl
<< "  -c list    comma-separated, 0-based list of columns to read" << endl
<< "  -e list    as above, columns to strip the quote character from" << endl
<< "  -E list    as above, columns to be enclosed with quote character" << endl
<< "  -q         quote character to be used with '-e'" << endl
<< "  -Q         quote character to be used with '-E'" << endl
<< "  -s         input content columns separator (',')" << endl
<< "  -S         output content columns separator (',')" << endl
<< "  --version  version info" << endl
<< endl
<< " Description:" << endl
<< " By default the program reads standard input and writes it back to standard"
<< endl
<< " output. This behaviour can be altered with the -i and -o switches followed by "
<< endl
<< " file name."
<< endl
<< " Note, when reading from standard input, Ctrl+D (Linux) or Ctrl+Z (Windows)"
<< endl
<< " can be used to send EOF character and break the loop."
<< endl
<< endl ;

}

