//
// scsv.cpp: libscsv library
//
// (C) Copyright 2016 TinyIT
// Written by Pawel Tatera
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public License
// as published by the Free Software Foundation; version 3
// of the License.
//


#include <iostream>		// i/o via console
#include <new>			// std::nothrow support
#include <fstream>		// i/o with files
#include <cstring>		// strlen, strcpy, memset, etc.
#include <list>
#include <vector>
#include <string>		// std::getline to read from istream

#include "scsv.h"		// needs: <list>,<vector>

#define CSV_COLS_CNT	16	// default cols count for CSV content; [1..N]


scsv::scsv(char iSeparator, char oSeparator,
  char quoteCh, char unquoteCh){
	this->init(iSeparator, oSeparator, quoteCh, unquoteCh);
	this->read_cols();
	this->quote_cols();
	this->unquote_cols();
	this->p.emptyLines=false;
	this->s = OK;
}


// Copy constructor and assignment operator.
scsv::scsv(const scsv &s){
	// Remove whatever is stored at the moment
	this->csv_free();
	this->copy(s);	
}


scsv & scsv::operator=(const scsv &s){
	// a=a assignment check, continue only if not the same
	if(this != &s){
		this->csv_free();
		this->copy(s);
	}

	return * this;
}


scsv::~scsv(){
	this->csv_free();
}


// Set defaults to be used.
void scsv::init(char iSeparator, char oSeparator,
  char quoteCh, char unquoteCh){
	this->p.iSeparator = iSeparator;
	this->p.oSeparator = oSeparator;
	this->p.quoteCh = quoteCh;
	this->p.unquoteCh = unquoteCh;
}


// Set columns to load via scsv::read.
// Call it without any parameters to reset.
void scsv::read_cols(unsigned short readColsL, unsigned short *readCols){
	if(readColsL == 0 || readCols == NULL){
		this->p.needColsL = 0;
		this->p.needCols = NULL;
	} else {
		this->p.needColsL = readColsL;
		this->p.needCols = readCols;
	}
}


// Set columns to be stripped from the quote character.
// Call it without any parameters to reset.
void scsv::quote_cols(unsigned short l, unsigned short *q){
	if(l == 0 || q == NULL){
		this->p.quoteColsL = 0;
		this->p.quoteCols = NULL;
	} else {
		this->p.quoteColsL = l;
		this->p.quoteCols = q;
	}
}


// Set columns to be enclosed with quote character.
// Call it without any parameters to reset.
void scsv::unquote_cols(unsigned short l, unsigned short *uq){
	if(l == 0 || uq == NULL){
		this->p.unquoteColsL = 0;
		this->p.unquoteCols = NULL;
	} else {
		this->p.unquoteColsL = l;
		this->p.unquoteCols = uq;
	}
}


// Read from file.
int scsv::read(const char *const fName, bool erase){
	std::ifstream f;

	if(erase) this->csv_free();

	if(fName == NULL) this->s = E_NULL;
	else if(*fName == 0) this->s = E_STR_EMPTY;
	else {
		f.open(fName);
		if(f.is_open()){
			this->read(f);
			// this->s already set
			f.close();
		} else this->s = E_READ_OPEN;
	}

	return this->s;
}


// Read from input stream (like cin)
// To send EOF and break the loop under Linux, use Ctrl+D.
// For Windows use Ctrl+Z.
int scsv::read(std::istream & is){
	std::string s;

	this->s = OK;

	while((this->s < E_OUT_OF_MEMORY) && std::getline(is, s))
		this->csv_add(this->csv_parse(s.c_str()));

	if(is.bad() || (is.fail() && !is.eof())) 
		this->s = E_READ_FAIL; 

	return this->s;
}


// Save content from memory into the file.
// if append set to true, append into existing file, overwrite otherwise.
int scsv::save(const char *const fName, bool append){
	std::ofstream of;
	std::ios_base::openmode om;

	if(fName == NULL) this->s = E_NULL;
	else if(*fName == 0) this->s = E_STR_EMPTY;
	else {
		om = std::ios_base::out; // text based output
		if(append) om |= std::ios_base::ate | std::ios_base::app ;

		of.open(fName, om);
		if(of.is_open()){
		  // this->s already set up
		  this->save(of);
		  of.close();
		} else this->s = E_SAVE_OPEN;
	}

	return this->s;
}


// Write the contents of memory into provided output stream (like console).
int scsv::save(std::ostream &os){
	std::list<std::vector<char *> *>::iterator iR;
	std::vector<char *>::iterator iC, iCLast, iCTop;
	unsigned int i;
	unsigned short idx;

	// Test if good bit is set
	if(os){
		this->s=OK;

		for(iR=this->csvContent.begin(); iR!=this->csvContent.end();
		  ++iR){
			// iR points at each underlying std::vector 
			// NULL equals empty line
			if(*iR){
				// iC is used to read CSV from iR-pointed
				// vectors.
				iCTop = (*iR)->end(); // behind last element 
				iCLast = iCTop - 1;
				i=0;
				for(iC = (*iR)->begin(); iC != iCTop; iC++){
					// test and write the quote char
					idx = scsv::map_column_idx(
						this->p.quoteCols,
						this->p.quoteColsL,
						i++);
					if(idx != 0xffff) os << this->p.quoteCh;

					// put iterated column content
					if(* iC) os << *iC;

					if(idx != 0xffff) os << this->p.quoteCh;

					// Put separator after every
					// column but not the last one.
					if(iC != iCLast)
						os << this->p.oSeparator;
				}
				os << std::endl;
			} else {
				// Write an empty line, if requested or
				// issue warning.
				if(this->p.emptyLines) os << std::endl;
				else this->s = W_STR_EMPTY;
			}
		}

	} else this->s = E_SAVE_FAIL;

	return this->s;
}


// Return 0-based row content.
int scsv::get(std::vector<char *> & rowV, size_t rowNum){

	const std::vector<char *> *v = this->csv_seek(rowNum); 
	if(v != NULL) rowV = *v;
	// else provided rowV remains unchanged

	return this->s;
}


//
// PRIVATE PART
//


// The broomstick...
void scsv::csv_free(){
	char *csvData;
	std::list<std::vector<char *> *>::iterator iR;
	std::vector<char *>::iterator iC, jC, iCTop;

	for(iR=this->csvContent.begin();
	  iR!=this->csvContent.end(); ++iR){
	  	// *iR: std::list entry (csv row)
		if(*iR){
			// iC: std::vector entry (csv column) 
			iCTop = (*iR)->end(); // behind last element 
			for(iC = (*iR)->begin(); iC != iCTop; iC++){
				csvData = *iC;
				if (csvData) delete [] csvData;

				// Zero out other vector columns pointing
				// at the same memory content...
				for(jC=iC+1; jC != iCTop; jC++) 
					if(*jC == csvData) *jC = NULL;
			}

			// Destroy std::vector content and the vector itself.
			(*iR)->clear();
			delete *iR;
		}
	}

	// Destroy std::list content and the vector itself.
	this->csvContent.clear();
	// TODO
	// When csvContent becomes HEAP-aligned (less overhead), then
	// delete this->csvContent; 
}


void scsv::copy(const scsv &x){
	// Copy known-size content;
	memcpy(& this->p, & x.p, sizeof(this->p));

	this->s = OK;

	// Copy unknown length/size, HEAP based content
	// this->csuvContent 
	this->csv_copy(x.csvContent);
}


void scsv::csv_copy(const std::list<std::vector<char *> *> & csvSrc){
	char *csvVal, *csvCpy;
	std::vector<char *> *rowP;

	std::list<std::vector<char *> *>::iterator iR;
	std::vector<char *>::iterator iC, jC;

	// Drop const-ness from the csvSrc; iterator isn't constant.
	for(iR=((std::list<std::vector<char *> *> &) csvSrc).begin();
	  (iR!=((std::list<std::vector<char *> *> &) csvSrc).end())
	  && (this->s == OK)  ; ++iR){
	  	// *iR: std::list entry (std::vector)
		if(*iR){
			// Create new row (vector)
			// this->p.needCols set from this::copy() level.
			rowP=this->row_create();
			if(! rowP) break;
			for(iC = (*iR)->begin(); iC != (*iR)->end(); iC++){
				csvVal = * iC;
				if(csvVal){
					// TODO
					// Rework to support column index-based
					// assignment. See csv_parse():435
					csvCpy = new char[strlen(csvVal)+1];
					if(csvCpy) strcpy(csvCpy, csvVal);
					else {
						this->s = E_OUT_OF_MEMORY; 
						break;
					}
				} else
					csvCpy = NULL; // empty column

				rowP->push_back(csvCpy);		

			}
		} else
			rowP = NULL;	

		// Append row to the container
		// NULL-pointers eq. empty lines.
		this->csv_add(rowP);
	}
}


// Tells if given c-string is quoted or not.
// If it is, the return value is the position of terminating quoting character
// location, compared to provided value.
const char * scsv::str_is_quoted(const char *str) const {
	const char *tStop;

	if((this->p.unquoteCh && str!=NULL) && *str==this->p.unquoteCh){
		tStop = strchr(++str, this->p.unquoteCh);
		if(tStop){
	// TODO
	// Bacsklash test infront of quoted character

	// If following character is not a NULL-terminating byte
	// or iSeparator, then we are parsing improperly quoted content here.
			if( *(tStop+1) != this->p.iSeparator &&
			*(tStop+1) != '\0' ) tStop = NULL;
		}
	} else tStop = NULL;

	return tStop;
}


// If not standing at the end of line, attempt to find next iSeparator.
const char * scsv::find_next_separator(const char * const lineP) const {
	return (*lineP=='\0')? lineP : strchr(lineP, this->p.iSeparator);
}


// Parse each csv line and build separate columns.
std::vector<char *> * scsv::csv_parse(const char * const csvLine){
	const char *t, *u;
	char *csvValue;
	unsigned short colIdx, tokIdx;
	union {
		unsigned int lineL;
		const char * csvLineEnd;
	};
	unsigned int tmpL;
	std::vector<char *> *csvRow;

	if(this->s >= E_OUT_OF_MEMORY) return NULL;

        lineL=strlen(csvLine);
	if(lineL==0){
		csvRow = NULL;
		if(! this->p.emptyLines) this->s = W_STR_EMPTY;
	} else
		csvRow = this->row_create(); // might return NULL too
		// but also set the p.s to error state.

	if(csvRow){
		csvLineEnd = csvLine + lineL;
		tokIdx=0;

		// Get 1st column token!
		t = csvLine;

		while(t <= csvLineEnd){
			// Determine where in 'this->p.needCols' should it stand.
			// If unset, copy entire row as is.
			if(this->p.needCols){
				colIdx=scsv::map_column_idx(this->p.needCols,
				  this->p.needColsL, tokIdx);
				// Is the column needed?
				if(colIdx == 0xffff){
					tokIdx++;
					t = this->find_next_separator(
						(const char *) t);
					if(t==NULL) t=csvLineEnd;
					else t++;
					continue;
				}
			} else {
				colIdx = tokIdx;
				// Need to resize the vector ourselves, since
				// the values are appended with 'operator[]',
				// not the 'push_back()'.
				if(colIdx >= csvRow->capacity()){
					csvRow->reserve(colIdx+CSV_COLS_CNT);
				}
			}

			// If not standing at the end of line attempt to find
			// next iSeparator.
			u = this->find_next_separator((const char *) t);

			if(u==t){
			// Empty columns: 'text;;text', ';text', 'text;'.
				csvValue=NULL; 
			} else {
			// Find next separator.
				if(u==NULL) u=csvLineEnd; 

				// Remove quote character from the column?
				register const char *q;
				q = str_is_quoted(t);
				if( q &&
				    scsv::map_column_idx(this->p.unquoteCols,
					this->p.unquoteColsL, colIdx)!=0xffff ){
				
					t++; // skip quote-start char
					tmpL = q-t; // skip quote-end char
				} else
					tmpL = u-t;

				csvValue = new (std::nothrow) char[tmpL+1];	
				if(csvValue==NULL){
					this->s=E_OUT_OF_MEMORY;
					break;
				}

				// TODO
				// Provide type recognition and conversion, 
				// Currently, only copy a c-string as is...
				strncpy(csvValue, t, tmpL);
				*(csvValue+tmpL)='\0';
			}

			if(this->p.needCols){
				// IMPORTANT! Index-based assignment.
				// DONT: csvRow->push_back(csvValue);
				// NOTE: we assign single csvValue to more than
				// one vector slot!
				while(colIdx != 0xffff){
				  (*csvRow)[colIdx] = csvValue;
				  colIdx=scsv::map_column_idx(this->p.needCols,
				    this->p.needColsL, tokIdx, colIdx+1);
				}
			} else {
				//(*csvRow)[colIdx] = csvValue;
				// The above won't work here.
				try {
					csvRow->push_back(csvValue);
				} catch (std::bad_alloc & ba){
					std::cerr << "bad_alloc caught: "
					<< ba.what() << std::endl
					<< "for item: " << csvValue
					<< std::endl;
					this->s=E_OUT_OF_MEMORY;
					break;
				}


			}

			// Loop for further columns
			tokIdx++;

			if(csvValue==NULL) t++;
			else t=u+1;
		}

	} // else this->s is already set.

	return csvRow;
}


// Add parsed row content (std::vector) to the this->csvContent (std::list).
// Function accepts NULL pointers on input - represent empty CSV line.
void scsv::csv_add(std::vector<char *> * csvRow){
	try {
		if(this->s < E_OUT_OF_MEMORY)
			this->csvContent.push_back(csvRow);

	} catch (std::bad_alloc & ba){
		std::cerr << "bad_alloc caught: "
		<< ba.what() << std::endl
		<< "For csvContent's row: " << this->csvContent.size()
		<< std::endl;
		this->s = E_OUT_OF_MEMORY;
	}
}


// Provides row of data from CSV, specified by the 0-based index number.
const std::vector<char *> * scsv::csv_seek(size_t rowNum){
	if(rowNum >= csvContent.size()){
		this->s = E_OUT_OF_RANGE;
		return NULL;
	} else {
		this->s = OK;
		std::list<std::vector<char *> *>::iterator iR =
		  csvContent.begin();

		// iR += rowNum; // doesn't compile(!?)
		// Walk around:
		for(size_t i=0; i<rowNum; i++, iR++);

		return *iR;
	}
}


std::vector<char *> * scsv::row_create() {
	std::vector<char *> *csvRow;

	// Note: p.needColsL = 0..N
	csvRow = new (std::nothrow) std::vector<char *>(this->p.needColsL);
	if(! csvRow)
		this->s = E_OUT_OF_MEMORY;
	else {
		if(this->p.needColsL==0)
			csvRow->reserve(CSV_COLS_CNT);
		// Bit overhead - keep space for default no. of columns.
		// this->s = OK; // consider
	}

	return csvRow;	
};




// Match columns from CSV file with ones required via read_cols.
unsigned short scsv::map_column_idx(unsigned short *list, unsigned short listL,
  unsigned short idx, unsigned short i) {
	unsigned short m = 0xffff;
	for(i; i<listL; i++){
		if(list[i] == idx){ m=i; break; }
	}

	return m;
}

